import click

import mucus.command


class Command(mucus.command.Command):
    def __call__(self, player, **kwargs):
        if player.events.loop.is_set():
            player.events.loop.clear()
        else:
            player.events.loop.set()
        click.echo(player.events.loop.is_set() and 'loop' or 'noop')



#!/usr/bin/env python3

import pathlib

import click

import mucus.command
import mucus.config
import mucus.deezer
import mucus.exception
import mucus.history
import mucus.player


@click.command()
@click.option('--config', '-c', 'config_path', type=click.Path(dir_okay=False, path_type=pathlib.Path), default=mucus.config.default_path())
@click.option('--version', is_flag=True)
def main(config_path, version):
    if version:
        click.echo(mucus.__version__)
        return

    config = mucus.config.load(config_path)
    client = mucus.deezer.client.Client()

    if client.data.user_id == 0:
        click.secho('not logged in', fg='red')

    player = mucus.player.Player(client)

    try:
        player.start()
    except mucus.exception.MucusException as e:
        raise click.ClickException(e)

    def inputs():
        while True:
            try:
                yield input(config.get('prompt', '\u1e43 '))
            except EOFError:
                break

    history = mucus.history.History()
    aliases = config.get('aliases', {})

    with history:
        for line in inputs():
            try:
                loader = mucus.command.Loader(line=line, aliases=aliases)
            except mucus.exception.NoSuchCommand:
                name = config.get('default_command', 'search')
                try:
                    loader = mucus.command.Loader(name=name, args=line)
                except mucus.exception.NoSuchCommand as e:
                    raise click.ClickException(e)

            context = {'client': client,
                       'command': {'line': line,
                                   'name': loader.name,
                                   'args': loader.args},
                       'config': config,
                       'history': history,
                       'player': player}

            runner = mucus.command.Runner(loader, context)

            try:
                runner()
            except mucus.exception.CommandException as e:
                click.echo(e)
            except mucus.exception.Exit:
                break



import mucus.command


class Command(mucus.command.Command):
    def __call__(self, client, command, **kwargs):
        if command['args']:
            client.auth.arl = command['args'].strip()
        r = client.post('deezer_userAutolog', json={
            'APP_NAME': 'Deezer',
            'APPLICATION_ID': 632384,
            'AUTHORIZE_UNLOGGED': True,
            'CHECKFORM': True,
            'DEVICE_TOKEN': '7610ab8f03d8fbe5ca8e57af3709f9e633ec234b502a842afe606724423b5d4c'
        })
        client.auth.jwt = r['jwt']



import argparse
import json
import shlex

import click

import mucus.command


class Command(mucus.command.Command):
    def __call__(self, command, player, **kwargs):
        if player.song is None:
            return

        parser = argparse.ArgumentParser(add_help=False, exit_on_error=False)
        parser.add_argument('-j', '--json', action='store_true')
        parser.add_argument('-v', '--verbose', action='store_true')
        try:
            args = parser.parse_args(shlex.split(command['args']))
        except argparse.ArgumentError as e:
            raise mucus.exception.CommandException(e)

        if args.json:
            click.echo(json.dumps(player.song, indent=2))
            return

        click.echo(' '.join([
            click.style(player.song['SNG_TITLE'], fg='blue'),
            click.style(player.song['ART_NAME'], fg='green'),
        ]))

        if args.verbose:
            def hms(s):
                r = [int(s)//60, int(s)%60]
                if r[0] > 59:
                    r.insert(0, r[0]//60)
                    r[1] %= 60
                return ':'.join((f'{i:02}' for i in r))
            click.secho(player.song['ALB_TITLE'], fg='cyan')
            click.secho(hms(player.song['DURATION']), fg='magenta')



import pathlib

import tomlkit


def default_path(filename='data.toml'):
    return pathlib.Path().home().joinpath('.local', 'share', 'mucus', filename)


def dump(doc: tomlkit.TOMLDocument, path: pathlib.Path = None):
    if path is None:
        try:
            path = pathlib.Path(doc['path'])
        except KeyError:
            path = default_path()
    path.parent.mkdir(parents=True, exist_ok=True)
    doc.remove('path')
    with path.open('w') as f:
        tomlkit.dump(doc, f)
    doc['path'] = str(path)


def load(path: pathlib.Path = None) -> tomlkit.TOMLDocument:
    if path is None:
        path = default_path()
    try:
        with path.open('r') as f:
            doc = tomlkit.load(f)
    except FileNotFoundError:
        doc = tomlkit.document()
    doc['path'] = str(path)
    return doc


def get(key, path: pathlib.Path = None):
    return load(path).get(key)


def set(key, value, path: pathlib.Path = None):
    doc = load(path)
    doc[key] = value
    dump(doc, path)


def hget(name, key, path: pathlib.Path = None):
    try:
        return load(path)[name][key]
    except KeyError:
        return None


def hset(name, key, value, path: pathlib.Path = None):
    doc = load(path)
    try:
        doc[name]
    except KeyError:
        doc[name] = {}
    doc[name][key] = value
    dump(doc, path)



import importlib.util
import re

import mucus.exception


class Command:
    def __call__(self, **kwargs):
        pass


class Loader:
    def __init__(self, line=None, name=None, args=None, aliases=None):
        if aliases is None:
            aliases = {}

        if line is not None:
            try:
                line = aliases[line]
            except KeyError:
                pass

            match = re.match(r'^\s*([A-Za-z0-9_-]+)(.*)', line)
            if match is None:
                raise mucus.exception.NoSuchCommand

            name = match.group(1).replace('-', '_')
            args = match.group(2).strip()

        if name is None:
            raise mucus.exception.NoSuchCommand

        try:
            name = aliases[name]
        except KeyError:
            pass

        try:
            self.mod = self.import_module('.' + name, 'mucus.commands')
        except ModuleNotFoundError as e:
            raise mucus.exception.NoSuchCommand(e)

        self.cmd = self.mod.Command()
        self.name = name
        self.args = args

    def import_module(self, name, package=None):
        spec = importlib.util.find_spec(name, package)
        if spec is None:
            raise ModuleNotFoundError(name)
        loader = importlib.util.LazyLoader(spec.loader)
        spec.loader = loader
        module = importlib.util.module_from_spec(spec)
        loader.exec_module(module)
        return module


class Runner:
    def __init__(self, loader, context):
        self.loader = loader
        self.context = context

    def __call__(self):
        return self.loader.cmd(**self.context)



import readline

import mucus.data


class History:
    def __init__(self, key='items'):
        self.key = key

    def __enter__(self):
        self.restore = [readline.get_history_item(i + 1) for i in range(readline.get_current_history_length())]
        readline.clear_history()
        data = mucus.data.hget('history', self.key)
        if data:
            for s in data:
                readline.add_history(s)
        return self

    def __exit__(self, *args):
        data = [readline.get_history_item(i + 1) for i in range(readline.get_current_history_length())]
        if data:
            mucus.data.hset('history', self.key, data)
        readline.clear_history()
        for s in self.restore:
            readline.add_history(s)



class Page:
    method = None
    params = None

    def __init__(self, client, **kwargs):
        self.data = client.post(self.method, json={**self.params, **kwargs})

    def __repr__(self):
        return f'<{self.__class__.__name__}>'


class Artist(Page):
    method = 'deezer.PageArtist'
    params = {'art_id': None, 'lang': 'en', 'tab': 0}

    def __iter__(self):
        def generator():
            for song in self.data['TOP']['data']:
                yield song
        return generator()


class Album(Page):
    method = 'deezer.PageAlbum'
    params = {'alb_id': None, 'lang': 'en', 'tab': 0, 'header': True}


class Track(Page):
    method = 'deezer.PageTrack'
    params = {'sng_id': None}


class Playlist(Page):
    method = 'deezer.PagePlaylist'
    params = {
        'nb': 2000,
        'start': 0,
        'playlist_id': None,
        'lang': 'en',
        'tab': 0,
        'tags': True,
        'header': True
    }

    def __iter__(self):
        for song in self.data['SONGS']['data']:
            yield song


class Search(Page):
    method = 'deezer.PageSearch'
    params = {
        'query': None,
        'start': 0,
        'nb': 10,
        'suggest': True,
        'artist_suggest': True,
        'top_tracks': True
    }

    def __iter__(self):
        for track in self.data.get('TRACKS', []):
            yield track


class SearchMusic(Page):
    method = 'search.music'
    params = {
        'query': None,
        'start': 0,
        'nb': 40,
        'filter': 'all',
        'output': 'TRACK'
    }

    def __iter__(self):
        for track in self.data:
            yield track



import argparse
import json
import shlex

import click

import mucus.command


class Command(mucus.command.Command):
    def __call__(self, client, command, **kwargs):
        parser = argparse.ArgumentParser(add_help=False, exit_on_error=False)
        parser.add_argument('-j', '--json', action='store_true')
        try:
            args = parser.parse_args(shlex.split(command['args']))
        except argparse.ArgumentError as e:
            raise mucus.exception.CommandException(e)

        if args.json:
            click.echo(json.dumps(client.data._data, indent=2))
        else:
            click.echo(client.data._data['USER']['EMAIL'])



class MucusException(Exception):
    pass


class Exit(MucusException):
    pass


class CommandException(MucusException):
    pass


class MucusError(MucusException):
    pass


class NoSuchCommand(MucusError):
    pass


class SoxNotInstalled(MucusError):
    pass


class DeezerError(MucusError):
    pass



class DeezerAuthError(DeezerError):
    pass


class DeezerStreamError(DeezerError):
    def __init__(self, errors):
        super().__init__(*[f'{e['code']}: {e['message']}' for e in errors])



class NoMedia(DeezerError):
    pass


class NoSource(DeezerError):
    pass



import pathlib

import tomlkit


def default_path(filename='config.toml'):
    path = pathlib.Path().home().joinpath('.config', 'mucus')
    if filename is not None:
        path = path.joinpath(filename)
    return path


def load(path: pathlib.Path = None) -> tomlkit.TOMLDocument:
    if path is None:
        path = default_path()
    try:
        with path.open('r') as f:
            doc = tomlkit.load(f)
    except FileNotFoundError:
        doc = tomlkit.document()
    doc.add('path', str(path))
    return doc


def dump(config: tomlkit.TOMLDocument, path: pathlib.Path = None):
    if path is None:
        try:
            path = pathlib.Path(config['path'])
        except KeyError:
            path = default_path()
    path.parent.mkdir(parents=True, exist_ok=True)
    config.remove('path')
    with path.open('w') as f:
        tomlkit.dump(config, f)



import argparse
import pathlib
import shlex

import click
import halo

import mucus.command
import mucus.config
import mucus.deezer.page
import mucus.deezer.stream



class Command(mucus.command.Command):
    def __call__(self, client, command, config, player, **kwargs):
        parser = argparse.ArgumentParser(exit_on_error=False)
        parser.add_argument('-a', '--album', action='store_true')
        parser.add_argument('-d', '--directory')
        args = parser.parse_args(shlex.split(command['args']))

        if args.directory is not None:
            try:
                config['download']
            except KeyError:
                config['download'] = {}
            config['download']['directory'] = args.directory
            mucus.config.dump(config)

        if player.song is None:
            return

        def download(song, album=None):
            stream = mucus.deezer.stream.stream(song, client.data.license_token)

            media = next(stream)
            if media is None:
                raise mucus.exception.NoMedia

            source = next(stream)
            if source is None:
                raise mucus.exception.NoSource

            root = pathlib.Path(config['download']['directory'])

            def _(s):
                return s.replace('/', '|')

            if album is None:
                path = root.joinpath(_(song['ART_NAME']), _(song['ALB_TITLE']))
            else:
                path = root.joinpath(_(album['ART_NAME']), _(album['ALB_TITLE']))
            path.mkdir(parents=True, exist_ok=True)
            ext = media['format'] == 'FLAC' and 'flac' or 'mp3'
            path = path.joinpath(f'{int(song['TRACK_NUMBER']):02} - {_(song['SNG_TITLE'])}.{ext}')

            with path.open('wb') as f:
                for chunk in stream:
                    f.write(chunk)

        spinner = halo.Halo()
        if args.album:
            page = mucus.deezer.page.Album(client=client, alb_id=player.song['ALB_ID'])
            for song in page.data['SONGS']['data']:
                spinner.start(f'{int(song['TRACK_NUMBER']):02} - {song['SNG_TITLE']}')
                try:
                    download(song, album=page.data['DATA'])
                except mucus.exception.MucusError:
                    spinner.fail()
                else:
                    spinner.succeed()
        else:
            with spinner:
                try:
                    download(player.song)
                except mucus.exception.MucusError:
                    spinner.fail()
                else:
                    spinner.succeed()



import hashlib
import random

import click
import halo

import mucus.command
import mucus.deezer.page
import mucus.history


class Command(mucus.command.Command):
    def __call__(self, client, command, player, **kwargs):
        if command['args'].strip() == '':
            return

        params = {'query': command['args'], 'nb': 20, 'start': 0}
        params_hash = hashlib.sha256(''.join([k + str(v) for k, v in params.items()]).encode()).hexdigest()[:16]

        def search():
            page = mucus.deezer.page.Search(client=client, **params)
            for k in page.data['ORDER']:
                results = page.data[k]
                if results:
                    try:
                        results = results['data']
                    except (KeyError, TypeError):
                        pass
                    for result in results:
                        t = result.get('__TYPE__')
                        if t == 'song':
                            yield result
                        elif t == 'artist':
                            artist = mucus.deezer.page.Artist(
                                client=client,
                                art_id=result['ART_ID']
                            )
                            for song in artist:
                                yield song
                        elif t == 'playlist':
                            playlist = mucus.deezer.page.Playlist(
                                client=client,
                                playlist_id=result['PLAYLIST_ID']
                            )
                            for song in playlist:
                                yield song

        def unique(songs):
            seen = set()
            for song in songs:
                if song['SNG_ID'] not in seen:
                    seen.add(song['SNG_ID'])
                    yield song

        songs = unique(search())

        while True:
            with halo.Halo():
                choices = []
                while len(choices) < 20:
                    try:
                        choices.append(next(songs))
                    except StopIteration:
                        break
                if len(choices) == 0:
                    break
            for i, track in enumerate(choices):
                click.echo(' '.join([
                    click.style(f'{i:>2}', fg='red'),
                    click.style(track['ART_NAME'], fg='green'),
                    click.style(track['SNG_TITLE'], fg='blue'),
                    click.style(track['ALB_TITLE'], fg='cyan')
                ]))
            with mucus.history.History(key=params_hash):
                try:
                    i = input('# ')
                except EOFError:
                    return
            shuffle = False
            if i == '.':
                continue
            elif ':' in i:
                if '*' in i:
                    shuffle = True
                    i = i.replace('*', '')
                i = slice(*map(lambda x: x.isdigit() and int(x) or None, i.split(':'))) # noqa
            else:
                try:
                    i = int(i)
                except ValueError:
                    return
                i = slice(i, i+1)
            songs = choices[i]
            if shuffle:
                random.shuffle(songs)
            for choice in songs:
                player.queue.put(choice)
            break



import httpx

import mucus.deezer.decrypter
import mucus.exception

default_formats = ('FLAC', 'MP3_320', 'MP3_128', 'MP3_64', 'MP3_MISC')


def stream(song, license_token, formats=default_formats):
    r = httpx.post('https://media.deezer.com/v1/get_url', json={
        'license_token': license_token,
        'media': [
            {
                'type': 'FULL', 'formats': [
                    {
                        'cipher': 'BF_CBC_STRIPE',
                        'format': fmt
                    } for fmt in formats
                ]
            }
        ],
        'track_tokens': [song['TRACK_TOKEN']]
    }).json()
    try:
        raise mucus.exception.DeezerStreamError(r['data'][0]['errors'])
    except (IndexError, KeyError):
        pass
    media = r['data'][0]['media'][0]
    yield media
    source = media['sources'][0]
    yield source
    decrypter = mucus.deezer.decrypter.Decrypter(song['SNG_ID'])
    pos = 0
    while True:
        try:
            with httpx.stream('GET', source['url'], headers={'range': f'bytes={pos}-'}, timeout=1.0) as r: # noqa
                for chunk in r.iter_bytes(chunk_size=2048):
                    if (pos // 2048) % 3 == 0 and len(chunk) == 2048:
                        chunk = decrypter.decrypt(chunk)
                    yield chunk
                    pos += len(chunk)
                return
        except httpx.HTTPError:
            continue



import argparse
import shlex

import click

import mucus.command
import mucus.config


class Command(mucus.command.Command):
    def __call__(self, command, config, **kwargs):
        try:
            config['aliases']
        except KeyError:
            config['aliases'] = {}

        if command['args'].strip() == '':
            for alias, command in config['aliases'].items():
                click.echo(' '.join([
                    click.style(alias, fg='red'),
                    '=',
                    click.style('"' + command +'"', fg='green')
                ]))
            return
        
        parser = argparse.ArgumentParser(exit_on_error=False)
        parser.add_argument('alias')
        parser.add_argument('command', nargs='*')
        try:
            args = parser.parse_args(shlex.split(command['args']))
        except argparse.ArgumentError as e:
            raise mucus.exception.CommandException(e)

        config['aliases'][args.alias] = shlex.join(args.command)
        mucus.config.dump(config)



import collections
import queue
import signal
import subprocess
import threading

import mucus.deezer.stream
import mucus.exception

events = ('play', 'stop', 'loop')


class Events(collections.namedtuple('Events', events)):
    def __new__(cls):
        return super().__new__(cls, *(threading.Event() for _ in events))


class Player:
    def __init__(self, client):
        self.client = client
        self.queue = queue.Queue()
        self.events = Events()
        self.thread = None
        self.song = None

    def loop(self):
        self.events.stop.clear()
        while not self.events.stop.is_set():
            try:
                song = self.queue.get(timeout=1)
            except queue.Empty:
                continue
            if song is None:
                self.queue.task_done()
                break
            else:
                self.play(song)
                self.queue.task_done()

    def play(self, song):
        while True:
            data = mucus.deezer.stream.stream(song, self.client.data.license_token)

            media = next(data)
            if media is None:
                raise mucus.exception.NoMedia

            source = next(data)
            if source is None:
                raise mucus.exception.NoSource

            self.song = song
            self.events.play.set()

            opts = ['--buffer', '2048', '-q', '-V0']
            if media['format'].startswith('MP3_'):
                opts.insert(0, '-t')
                opts.insert(1, 'mp3')

            with subprocess.Popen(['sox', *opts, '-', '-d'], bufsize=2048, stdin=subprocess.PIPE) as p:
                for chunk in data:
                    if self.events.stop.is_set():
                        break

                    if not self.events.play.is_set():
                        p.send_signal(signal.SIGSTOP)
                        self.events.play.wait()
                        p.send_signal(signal.SIGCONT)

                    if self.events.stop.is_set():
                        break

                    p.stdin.write(chunk)

                p.stdin.close()

                if self.events.stop.is_set():
                    p.terminate()

            if self.events.stop.is_set():
                break
            elif self.events.loop.is_set():
                continue
            else:
                break
        self.song = None
        self.events.play.clear()

    def start(self):
        try:
            subprocess.run(['sox', '--version'], stdout=subprocess.DEVNULL)
        except FileNotFoundError:
            raise mucus.exception.SoxNotInstalled

        self.thread = threading.Thread(target=self.loop, daemon=True)
        self.thread.start()



import random

import httpx

import mucus.data
import mucus.exception


class Auth(httpx.Auth):
    def __init__(self, sid=None, arl=None):
        if sid is None or arl is None:
            data = mucus.data.get('auth')
            if data is not None:
                sid = data.get('sid')
                arl = data.get('arl')
        self.sid = sid
        self.arl = arl
        self.jwt = None

    def auth_flow(self, request):
        cookies = {}
        if self.arl is not None:
            cookies['arl'] = self.arl
        if self.sid is not None:
            cookies['sid'] = self.sid
        if cookies:
            request.headers['Cookie'] = '; '.join(f'{k}={v}' for k, v in cookies.items())
        response = yield request
        try:
            response.cookies['arl']
            response.cookies['sid']
        except KeyError:
            pass
        else:
            mucus.data.set('auth', {'sid': response.cookies['sid'],
                                    'arl': response.cookies['arl']})


class Client(httpx.Client):
    base_url = httpx.URL('https://www.deezer.com/ajax/')

    def __init__(self, auth=None):
        if auth is None:
            auth = Auth()

        super().__init__(auth=auth)

        self.data = UserData(self)

    def request(self, method, url, **kwargs):
        if url == 'deezer.getUserData':
            api_token = ''
        else:
            api_token = self.data.api_token

        kwargs['params'] = {
            'method': url,
            'input': 3,
            'api_version': '1.0',
            'api_token': api_token,
            'cid': random.randrange(0, 1000000000)
        }

        r = super().request(method, 'gw-light.php', **kwargs).json()
        if r.get('error'):
            raise mucus.exception.DeezerError(r['error'])

        return r['results']


class UserData:
    def __init__(self, client):
        self._data = client.post('deezer.getUserData')

    @property
    def api_token(self):
        return self._data['checkForm']

    @property
    def license_token(self):
        return self._data['USER']['OPTIONS']['license_token']

    @property
    def user_id(self):
        return self._data['USER']['USER_ID']
